# -*- coding: utf-8 -*-
"""
(c)2012 ЗАО "РИПАС" и Демин Федор,
www.ripas.ru dff@ripas.ru
"""
"""
Скрипт для парсинга протокола "TSpider"
"""

from PythonQt import *
from RoadCentre import kernel, plugins
from menu import Menu
import gui
import os
import metarpc

areaPStates = (
    u'Нет предопределённого состояния',
    u'Локальный режим (не управляется сервисом)',
    u'Жёлтое мигание',
    u'Отключенное состояние'
)

controlSources = (
    u'не управляется службой',
    u'календарная автоматика',
    u'адаптивный модуль',
    u'локальное управление',
    u'оператор'
)

def __paramToStr(param, n):
    if n < 0: return u'отсутствует'
    if n < len(param): return param[n]
    return u"неизвестный тип"

def updateZoneToolTip(obj):
    if hasattr(obj,'hashInvalid'):
        z_menu=Menu.devClassMenus['TSpider::Zone']
        if z_menu:
            z_menu.disabled=obj.hashInvalid
        a_menu=Menu.devClassMenus['TSpider::Area']
        if a_menu:
            a_menu.disabled=obj.hashInvalid
        if obj.hashInvalid:
            obj.toolTip=u'<font color="#ff0000">Локальная конфигурация не совпадает с конфигурацией сервера. Необходимо обновить проект!</font>'
            return

    toolTip = '<nobr>' \
        u'Зона управления: <b>%u - %s</b><hr>' % (obj.id, obj.name);

    if not obj.isConnected:
        toolTip += u"Отключена" u"<br>Последний раз была на связи <b>"
        if obj.lastConnectionEventTime.isValid():
            toolTip += obj.lastConnectionEventTime.toString(QtCore.Qt.DefaultLocaleShortDate) + "</b>"
        else:
            toolTip += u"на связи не была</b>"

    else:
        currentStrategy = obj.currentStrategy
        if currentStrategy > 0:
            strategy = obj.strategy(currentStrategy)
            if strategy: stratName = strategy.name
            else: stratName = u'неизвестная'
            stratName = '%i - %s' % (currentStrategy, stratName)
        else: stratName = u'отсутствует'
        toolTip += \
            u'Подключена: <b>%s</b><br>' \
            u'Источник управления: <b>%s</b><br>' \
            u'Стратегия: <b>%s</b>' \
            % (
                obj.connectionTime.toString(QtCore.Qt.DefaultLocaleShortDate),
                __paramToStr(controlSources, obj.controlSource),
                stratName
            )

    obj.toolTip=toolTip + '</nobr>'

def updateAreaToolTip(obj):
    if obj.specialState:
        state = __paramToStr(areaPStates, obj.specialState)
    else:
        state = '%i - ' % obj.currentState
        s = obj.state(obj.currentState)
        if s: state = state + s.name
        else: state = state + u'нет в списке'
    z = obj.parent().parent()

    obj.toolTip = '<nobr>' \
        u'Район управления: <b>%u - %s</b><br>' \
        u'Зона: <b>%u - %s</b><hr>' \
        u'Источник управления: <b>%s</b><br>' \
        u'Режим: <b>%s</b><br>' \
        '</nobr>' \
        % (obj.id, obj.name, z.id, z.name, __paramToStr(controlSources, obj.controlSource), state)

def onConnectionMsg(msg):
    if msg['protocol'] == 0x29111992:
        data = msg.pop('data')
        if data: data = data.toHex()
        gui.msgInfo(u'TSpider connection message:' + str(msg) + str(data))

def zoneRequest(request, zone):
    gui.msgInfo(u"Зона:%u(%s). Отправлена команда: %s"%(zone.id, zone.name, request.toString()))

def zoneResponseFinished(response, zone):
    gui.msgInfo(u"Зона:%u(%s). Ответ на команду: %s. Результат: %s %s"%(zone.id, zone.name, response.request.toString(), response.errorString(True), response.toString()))

def zoneNotification(notification, zone):
    gui.msgInfo(u"Зона:%u(%s). Сообщение тип:%s"%(zone.id, zone.name, notification.toString()))

def loadZone(fileName):
    z = plugins.TSpiderPlugin.load(fileName)
    if z:
        from RoadCentre import project
        z.findChildren('QtCore.QObject')
        kernel.itemModel.addObjects(u'Зоны',(z,))
        z.setConnection(project.connection)
        z.connect('request(const QObject*, QObject*)', zoneRequest)
        z.connect('responseFinished(const QObject*, QObject*)', zoneResponseFinished)
        z.connect('notification(const QObject*, QObject*)', zoneNotification)
        z.setScene(kernel.graphicsScene)
        z.connect('stateChanged(const QObject*)', updateZoneToolTip)
        z.connect('stateChanged(const QObject*)', gui.selectionTable.objectStateChanged)
        z.connect('areaStateChanged(const QObject*)', updateAreaToolTip)
        z.connect('areaStateChanged(const QObject*)', gui.selectionTable.objectStateChanged)
        updateZoneToolTip(z)
        layer=kernel.graphicsScene.addLayer(u'Зоны')
        layer.addItem(z.graphicsItem())
        z.refresh()
    return z

def loadZones(dir_name):
    if not hasattr(plugins,'TSpiderPlugin'): return
    import os
    zones_path=os.path.join(dir_name,'tspider')
    for dirpath, dirnames, filenames in os.walk(zones_path):
        for name in filenames:
            if os.path.splitext(name)[-1] != '.xtsc': continue
            loadZone(os.path.join(dirpath, name))

def reloadZone(zone_file):
    for z in plugins.TSpiderPlugin.findChildren('TSpider::Zone'):
        if z.modelData(2,QtCore.Qt.DisplayRole)==zone_file:
            z.deleteLater()
            loadZone(zone_file)
            return
    loadZone(zone_file)

def createZone():
    if not hasattr(plugins,'TSpiderPlugin'):
        QtGui.QMessageBox.critical(gui.mainWindow,u'Ошибка',u'Нет модуля TSpider')
        return
    cnfStorage=None
    if hasattr(metarpc,'cnfStorage'):
        cnfStorage=metarpc.cnfStorage.rpc
    from RoadCentre import project
    plugins.TSpiderPlugin.editZone(os.path.join(project.baseDir,'tspider'),cnfStorage)

def configBlockFinished():
    if not configBlockFinished.resp.data or configBlockFinished.resp.data[0].isEmpty():
        downloadZone.load_wnd.accept()
        return
    size=configBlockFinished.resp.data[0].size()
    downloadZone.progressBar.value+=size
    downloadZone.conf.append(configBlockFinished.resp.data[0])
    configBlockFinished.resp=downloadZone.z.sendCommand('GET_CONFIG',downloadZone.progressBar.value)
    configBlockFinished.resp.connect('finished()',configBlockFinished)

def configSizeFinished():
    downloadZone.progressBar.setRange(0,configSizeFinished.resp.data[0])
    configBlockFinished.resp=downloadZone.z.sendCommand('GET_CONFIG',downloadZone.progressBar.value)
    configBlockFinished.resp.connect('finished()',configBlockFinished)

def downloadZoneFinished(result):
    if result==QtGui.QDialog.Accepted:
        zone_file=downloadZone.z.modelData(2,QtCore.Qt.DisplayRole)
        with open(zone_file,'wb+') as zf:
            zf.write(downloadZone.conf.data())
        reloadZone(zone_file)

class ZoneUploader:
    def __init__(self,zone):
        self.zone = zone
        self.config = None
        self.load_wnd = None
        self.progress_bar = None
        self.block_size = None
        self.zone_id = None
        self.lock_id = None

    def start(self):
        if not self.zone.inherits('TSpider::Zone'):
            return self._show_error('Неверный объект зоны управления')
        f=QtCore.QFile(self.zone.modelData(2,0))
        if not f.open(QtCore.QIODevice.ReadOnly):
            return self._show_error('Невозможно открыть файл конфигурации зоны управления')
        self.config = f.readAll()
        f.close()
        self.load_wnd=gui.widgetFromUi('zoneloading.ui')
        if self.load_wnd is None:
            return self._show_error('Ошибка при создании окна загрузки')
        self.progress_bar=self.load_wnd.findChild('QProgressBar','progressBar')
        if self.progress_bar is None:
            return self._show_error('Неверный экземпляр окна загрузки')
        self.zone_id = self.zone.id
        self.progress_bar.setRange(0,self.config.size())
        self.progress_bar.value = 0

        self.load_wnd.connect('rejected()',self._upload_cancelled)

        resp = self.zone.sendCommand('LOCK_CONFIG')
        resp.connect('responseFinished(Response*)',self._lock_response_finished)

        getattr(self.load_wnd,'exec')()

    def _upload_cancelled(self):
        self.zone = None

    def _lock_response_finished(self,resp):
        if resp.error != 0:
            self._show_error(resp.errorString(True))
            self.load_wnd.accept()
            return
        self.lock_id = resp.data[0]
        self._upload_next_block()

    def _upload_next_block(self):
        if self.zone is None:
            return
        self.block_size=1024
        if self.progress_bar.value+self.block_size > self.progress_bar.maximum:
            self.block_size = self.progress_bar.maximum - self.progress_bar.value
        if self.block_size <= 0:
            resp = self.zone.sendCommand('APPLY_CONFIG',(self.lock_id,))
            resp.connect('responseFinished(Response*)',self._apply_config_finished)
            self.zone.connect('connected(bool)',self._zone_connected)
            return
        resp = self.zone.sendCommand('SET_CONFIG',(self.lock_id,self.progress_bar.value,self.config.mid(self.progress_bar.value,self.block_size)))
        resp.connect('responseFinished(Response*)',self._response_finished)

    def _zone_connected(self,conn_state):
        if not conn_state:
            QtCore.QTimer.singleShot(1000,self._refresh_zone)
            self.load_wnd.accept()

    def _response_finished(self, resp):
        resp.deleteLater()
        if resp.error != 0x1200:
            self._show_error(resp.errorString(True))
            self.load_wnd.accept()
            return
        self.progress_bar.value += self.block_size
        self._upload_next_block()

    def _apply_config_finished(self, resp):
        resp.deleteLater()
        if self.zone is None:
            return
        if resp.error != 0x1200:
            self._show_error('Ошибка при применении конфигурации: ' + resp.errorString(True))
        else:
            QtCore.QTimer.singleShot(1000,self._refresh_zone)
        self.load_wnd.accept()

    def _refresh_zone(self):
        for z in plugins.TSpiderPlugin.findChildren('TSpider::Zone'):
            if z.id == self.zone_id:
                z.refresh()
                break

    def _show_error(self,msg):
        QtGui.QMessageBox.critical(gui.mainWindow,u'Ошибка',msg)


def downloadZone(z):
    if not z.inherits('TSpider::Zone'):
        return
    import gui
    downloadZone.load_wnd=gui.widgetFromUi('zoneloading.ui')
    del gui
    if not downloadZone.load_wnd:
        return
    downloadZone.progressBar=downloadZone.load_wnd.findChild('QProgressBar','progressBar')
    if not downloadZone.progressBar:
        return
    downloadZone.progressBar.value=0
    downloadZone.z=z
    downloadZone.conf=QtCore.QByteArray()
    configSizeFinished.resp=z.sendCommand('GET_CONFIG_SIZE')
    configSizeFinished.resp.connect('finished()',configSizeFinished)
    downloadZone.load_wnd.connect('finished(int)',downloadZoneFinished)
    getattr(downloadZone.load_wnd,'exec')()

def uploadZone(z):
    uploader = ZoneUploader(z)
    uploader.start()

from menu import SubMenu

class ZoneMenu(SubMenu):
    def __init__(self):
        SubMenu.__init__(self, (u'Зона', u'Зоны'))
        self.menuStrategy = Qt.QMenu(u'Установить стратегию')
        gr = Qt.QActionGroup(None)
        self.__grStrategy = gr
        gr.connect('triggered(QAction*)', self.actionCommand)

        gr = Qt.QActionGroup(None)
        self.__grCmd = gr
        gr.connect('triggered(QAction*)', self.actionCommand)
        a = self.menuMain.addAction(u'Отключить управление зоной')
        a.setData(('SET_LOCAL', 0))
        gr.addAction(a)
        a = self.menuMain.addAction(u'Установить режим по-умолчанию')
        a.setData(('SET_DEFAULT', 0))
        gr.addAction(a)
        a = self.menuMain.addAction(u'Скрыть/показать')
        a.connect('triggered()', self.actionShow)
        self.downloadZoneAction = self.menuMain.addAction(u'Загрузить с сервера')
        self.downloadZoneAction.connect('triggered()', self.downloadZone)
        self.downloadZoneAction.setData(('GET_CONFIG_SIZE','GET_CONFIG'))

        self.uploadZoneAction = self.menuMain.addAction(u'Загрузить на сервер')
        self.uploadZoneAction.connect('triggered()', self.uploadZone)

        if hasattr(plugins.TSpiderPlugin,'edit') and plugins.TSpiderPlugin.edit:
            a = self.menuMain.addAction(u'Изменить')
            a.connect('triggered()', self.editZone)
        self.menuMain.addSeparator()
        a = self.menuMain.addAction(u'Удалить')
        a.connect('triggered()', self.deleteZone)
        self.disabled=False
        self.edited_zone = None

    def actionCommand(self,act):
        data = act.data()
        cmd = data[0]
        param = data[1]
        for z in self.__connected:
            z.sendCommand(cmd, param)

    def actionShow(self):
        for o in self.selection:
            o.visible = not o.visible

    def downloadZone(self):
        for o in self.selection:
            downloadZone(o)

    def uploadZone(self):
        for o in self.selection:
            uploadZone(o)

    def deleteZone(self):
        from PythonQt import QtGui
        from gui import mainWindow
        import os
        if QtGui.QMessageBox.question(mainWindow,u'Вопрос',u'Удалить выбранные зоны?',QtGui.QMessageBox.Yes|QtGui.QMessageBox.No)!=QtGui.QMessageBox.Yes:
            return
        for o in self.selection:
            if not o.inherits('TSpider::Zone'):
                continue
            os.remove(o.modelData(2,QtCore.Qt.DisplayRole))
            o.deleteLater()

    def menu(self, obj):
        if not len(obj): return None

        connected = []       #Список подключенных
        for o in obj:
            if o.isConnected:
                connected.append(o)
        self.__connected = connected
        count = len(connected)
        self.__grCmd.visible = count != 0
        m = self.menuStrategy
        m.clear()
        m.enabled = count == 1
        self.downloadZoneAction.enabled = count == 1
        self.uploadZoneAction.enabled = count == 1
        if self.disabled:
            m.enabled=False
        #m.visible = count == 1
        if count == 0:
            return SubMenu.menu(self, obj)
        if count == 1:
            o = obj[0]
            self.downloadZoneAction.enabled=o.connected
            self.uploadZoneAction.enabled=o.connected
            if not o.isConnected: return None
            gr = self.__grStrategy
            for s in o.strategies():
                if not hasattr(s,'id'): continue
                a = m.addAction('%i:%s'%(s.id, s.name))
                a.setData(('SET_STRAT',(s.id,0)))
                if o.currentStrategy == s.id:
                    a.checkable = True
                    a.checked = True
                gr.addAction(a)
            m.enabled = not self.disabled

        no_menu_strategy=True
        try:
            a=m.actions()[0];
            a.visible=SubMenu.checkAction(a,'TSpider')
            if a.visible: no_menu_strategy=False
        except: pass

        for a in self.menuMain.actions():
            a.visible=SubMenu.checkAction(a,'TSpider')

        if not no_menu_strategy:
            self.menuMain.insertMenu(self.menuMain.actions()[0],self.menuStrategy)
        return SubMenu.menu(self, obj)

    def editZone(self):
        from RoadCentre import project
        cnfStorage=None
        if hasattr(plugins,'MetaRpcPlugin'):
            cnfStorage=plugins.MetaRpcPlugin.findRpcObject('cnfStorage')
        for o in self.selection:
            plugins.TSpiderPlugin.editZone(o.modelData(2,QtCore.Qt.DisplayRole),cnfStorage)
            edit_wnd = gui.mainWindow.findChild('ZoneEditDlg')
            if edit_wnd:
                self.edited_zone = o
                edit_wnd.connect('finished(int)',self._zone_edit_finished)
            break

    def _zone_edit_finished(self,code):
        if code == QtGui.QDialog.Accepted and self.edited_zone is not None:
            ans = QtGui.QMessageBox.question(gui.mainWindow,u'Вопрос',u'Загрузить изменённую конфигурацию на сервер?',QtGui.QMessageBox.Yes|QtGui.QMessageBox.No)
            if ans == QtGui.QMessageBox.Yes:
                uploadZone(self.edited_zone)
        self.edited_zone = None


class AreaMenu(SubMenu):
    def __init__(self,kernel):
        SubMenu.__init__(self, (u'Район управления', u'Районы управления'))
        self.kernel = kernel

        self.menuConst = Qt.QMenu()
        a = self.menuConst.addAction(u'Скрыть/показать')
        a.connect('triggered()', self.actionShow)

        menuPState = Qt.QMenu(u'Установить особое состояние')
        self.__menuPState = menuPState
        gr = Qt.QActionGroup(None)
        self.__grPState = gr
        gr.connect('triggered(QAction*)', self.actionCommand)
        a = menuPState.addAction(u'По-умолчанию')
        a.setData(('SET_ADEFAULT',()))
        gr.addAction(a)
        a = menuPState.addAction(u'Локальный режим')
        a.setData(('SET_APSTATE',(1,0)))
        gr.addAction(a)
        a = menuPState.addAction(u'ЖМ')
        a.setData(('SET_APSTATE',(2,0)))
        gr.addAction(a)
        a = menuPState.addAction(u'ОС')
        a.setData(('SET_APSTATE',(3,0)))
        gr.addAction(a)
        self.menuConst.addMenu(menuPState)

        self.menuState = Qt.QMenu(u'Установить режим')
        gr = Qt.QActionGroup(None)
        self.__grState = gr
        gr.connect('triggered(QAction*)', self.actionSetState)
        self.disabled=False

    def actionCommand(self,act):
        data = act.data()
        cmd = data[0]
        param = data[1]
        for a in self.selection:
            self.zone(a).sendCommand(cmd, (a.id,) + param)

    def zone(self,area):
        zone=area.parent()
        while zone and not zone.inherits('TSpider::Zone'):
            zone=zone.parent()
        return zone

    #def actionSetPState(self,act):
        #state = act.data()
        #for a in self.selection:
            #self.self.zone(a).sendCommand('SET_APSTATE', (a.id, state, 0))

    def actionSetState(self,act):
        state = act.data()
        for a in self.selection:
            self.zone(a).sendCommand('SET_ASTATE', (a.id, state, 0))

    def actionShow(self):
        for o in self.selection:
            o.visible = not o.visible

    def menu(self, obj):
        count = len(obj)
        if not count or self.disabled: return None
        self.menuMain.clear()
        self.menuMain.addActions(self.menuConst.actions())
        if count == 1:
            menuState = self.menuState
            menuState.clear()
            self.menuMain.addMenu(self.menuState)
            gr = self.__grState

            o = obj[0]
            for s in o.states():
                a = menuState.addAction('%i:%s'%(s.id, s.name))
                a.setData(s.id)
                if o.currentState == s.id:
                    a.checkable = True
                    a.checked = True
                gr.addAction(a)
            act = self.__grPState.actions()
            for i in range(len(act)):
                a = act[i]
                ch = (i == o.specialState) and (i > 0)
                a.checkable = ch
                a.checked = ch

        return SubMenu.menu(self, obj)


